﻿using System;
using System.Collections.Generic;

namespace shuzihuarongdao
{
    public enum DifficultyLevel : int
    {
        EASY = 30,
        NORMAL = 60,
        HARD = 90,
    };

    public enum Direction : int
    {
        UP = -4,
        DOWN = 4,
        LEFT = -1,
        RIGHT = 1,
    };

    public class Board
    {
        // stores init number label on each block position.
        private List<int> initState = new List<int> { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 0 };
        // stores current number label on each block position.
        private List<int> currentState = new List<int> { };
        private int emptyBlockPosition = 15;

        public static List<Direction> ALL_DIRECTIONS = new List<Direction> { Direction.UP, Direction.LEFT, Direction.DOWN, Direction.RIGHT };
        private static Random RAND = new Random();

        public Board() => currentState = initState;

        public int GetPositionLabel(int position)
        {
            return currentState[position];
        }

        /**
         * Return true if given block position is on board
         * @param position should be 0 to 15, meaning each block position on the board.
         */
        private bool PositionOnBoard(int position)
        {
            return position >= 0 && position <= 15;
        }

        /**
         * create a new game board
         */
        public static Board GenerateBoard(DifficultyLevel difficultyLevel)
        {
            Board board = new Board();
            // empty block move path
            List<Direction> path = new List<Direction> { };
            while (true) {
                List<Direction> directions = board.GetEmptyBlockDirections();
                if (path.Count > 0)
                {
                    directions.Remove(GetOppositeDirection(path[path.Count - 1]));
                }
                if (directions.Count == 0)
                {
                    Console.WriteLine("should not happen");
                    break;
                }
                Direction randomDirection = randomChoice(directions);
                board.MoveEmptyBlock(randomDirection);
                path.Add(randomDirection);
                if (path.Count >= (int)difficultyLevel)
                {
                    break;
                }
            }
            board.initState = board.currentState;
            //ShowPath(path);
            //ShowSolution(path);
            return board;
        }

        private static void ShowPath(List<Direction> path)
        {
            Console.WriteLine("Path: ");
            int i = 0;
            foreach (Direction d in path)
            {
                Console.Write(DirectionToString(d) + " ");
                i++;
                if (i % 10 == 0)
                {
                    Console.WriteLine();
                }
            }
            if (i % 10 != 0)
            {
                Console.WriteLine();
            }
        }

        /**
         * show solution, a list of empty block moving directions.
         */
        private static void ShowSolution(List<Direction> path)
        {
            Console.WriteLine("Solution: ");
            List<Direction> reversedPath = path;
            reversedPath.Reverse();
            int i = 0;
            foreach (Direction d in reversedPath)
            {
                Console.Write(DirectionToString(GetOppositeDirection(d)) + " ");
                i++;
                if (i % 10 == 0)
                {
                    Console.WriteLine();
                }
            }
            if (i % 10 != 0)
            {
                Console.WriteLine();
            }
        }

        private static string DirectionToString(Direction d)
        {
            switch (d)
            {
                case Direction.UP:
                    return "UP   ";
                case Direction.DOWN:
                    return "DOWN ";
                case Direction.LEFT:
                    return "LEFT "; 
                case Direction.RIGHT:
                    return "RIGHT";
                default:
                    return "";  // should not reach
            }
        }

        /**
         * return true if board is valid (not neccesarily solvable).
         */
        public bool IsValid()
        {
            // 0 to 15 appear exactly once
            bool[] appeared = new bool[16];
            foreach (int label in currentState)
            {
                if (appeared[label])
                {
                    return false;
                }
                else
                {
                    appeared[label] = true;
                }
            }
            foreach (bool b in appeared)
            {
                if (! b)
                {
                    return false;
                }
            }
            return true;
        }

        /**
         * return true if game is solved.
         */
        public bool IsSolved()
        {
            if (currentState[15] != 0)
            {
                return false;
            }
            for (int i = 0; i < 15; i++)
            {
                if (currentState[i] != i + 1)
                {
                    return false;
                }
            }
            return true;
        }

        /**
         * move empty block in given directions.
         */
        private void MoveEmptyBlock(Direction direction)
        {
            int newEmptyBlockPosition = emptyBlockPosition + (int)direction;
            currentState[emptyBlockPosition] = currentState[newEmptyBlockPosition];
            currentState[newEmptyBlockPosition] = 0;
            emptyBlockPosition = newEmptyBlockPosition;
        }

        private static Direction randomChoice(List<Direction> directions)
        {
            return directions[RAND.Next(directions.Count)];
        }

        private static Direction GetOppositeDirection(Direction direction)
        {
            switch (direction)
            {
                case Direction.UP:
                    return Direction.DOWN;
                case Direction.DOWN:
                    return Direction.UP;
                case Direction.LEFT:
                    return Direction.RIGHT;
                case Direction.RIGHT:
                    return Direction.LEFT;
                default:
                    return Direction.DOWN;  // should not happen
            }
        }

        /**
         * get directions that empty block can move.
         */
        private List<Direction> GetEmptyBlockDirections()
        {
            List<Direction> directions = new List<Direction> { };
            foreach (Direction direction in ALL_DIRECTIONS)
            {
                if (PositionOnBoard(emptyBlockPosition + (int)direction))
                {
                    if ((direction == Direction.UP && emptyBlockPosition > 3) ||
                        (direction == Direction.DOWN && emptyBlockPosition < 12) ||
                        (direction == Direction.LEFT && emptyBlockPosition % 4 != 0) ||
                        (direction == Direction.RIGHT && emptyBlockPosition % 4 != 3))
                    {
                        directions.Add(direction);
                    }
                }
            }
            return directions;
        }

        /**
         * Return true if given block can move.
         * @param position should be 0 to 15, meaning each block position on the board.
         */
        public bool BlockCanMove(int position)
        {
            if (currentState[position] == 0)
            {
                return false;
            }
            var indexToCheck = new List<int> { position - 4, position + 4, position - 1, position + 1 };
            foreach (var index in indexToCheck)
            {
                if (index == emptyBlockPosition)
                {
                    return true;
                }
            }
            return false;
        }

        public int GetEmptyBlockPosition()
        {
            return emptyBlockPosition;
        }

        public void MoveBlock(int position)
        {
            if (!BlockCanMove(position))
            {
                return;
            }
            currentState[emptyBlockPosition] = currentState[position];
            currentState[position] = 0;
            emptyBlockPosition = position;
        }
    }
}